#line 1 "scanner.cpp"

#line 3 "scanner.cpp"

#define  YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 6
#define YY_FLEX_SUBMINOR_VERSION 4
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

#ifdef yy_create_buffer
#define InstrScript__create_buffer_ALREADY_DEFINED
#else
#define yy_create_buffer InstrScript__create_buffer
#endif

#ifdef yy_delete_buffer
#define InstrScript__delete_buffer_ALREADY_DEFINED
#else
#define yy_delete_buffer InstrScript__delete_buffer
#endif

#ifdef yy_scan_buffer
#define InstrScript__scan_buffer_ALREADY_DEFINED
#else
#define yy_scan_buffer InstrScript__scan_buffer
#endif

#ifdef yy_scan_string
#define InstrScript__scan_string_ALREADY_DEFINED
#else
#define yy_scan_string InstrScript__scan_string
#endif

#ifdef yy_scan_bytes
#define InstrScript__scan_bytes_ALREADY_DEFINED
#else
#define yy_scan_bytes InstrScript__scan_bytes
#endif

#ifdef yy_init_buffer
#define InstrScript__init_buffer_ALREADY_DEFINED
#else
#define yy_init_buffer InstrScript__init_buffer
#endif

#ifdef yy_flush_buffer
#define InstrScript__flush_buffer_ALREADY_DEFINED
#else
#define yy_flush_buffer InstrScript__flush_buffer
#endif

#ifdef yy_load_buffer_state
#define InstrScript__load_buffer_state_ALREADY_DEFINED
#else
#define yy_load_buffer_state InstrScript__load_buffer_state
#endif

#ifdef yy_switch_to_buffer
#define InstrScript__switch_to_buffer_ALREADY_DEFINED
#else
#define yy_switch_to_buffer InstrScript__switch_to_buffer
#endif

#ifdef yypush_buffer_state
#define InstrScript_push_buffer_state_ALREADY_DEFINED
#else
#define yypush_buffer_state InstrScript_push_buffer_state
#endif

#ifdef yypop_buffer_state
#define InstrScript_pop_buffer_state_ALREADY_DEFINED
#else
#define yypop_buffer_state InstrScript_pop_buffer_state
#endif

#ifdef yyensure_buffer_stack
#define InstrScript_ensure_buffer_stack_ALREADY_DEFINED
#else
#define yyensure_buffer_stack InstrScript_ensure_buffer_stack
#endif

#ifdef yylex
#define InstrScript_lex_ALREADY_DEFINED
#else
#define yylex InstrScript_lex
#endif

#ifdef yyrestart
#define InstrScript_restart_ALREADY_DEFINED
#else
#define yyrestart InstrScript_restart
#endif

#ifdef yylex_init
#define InstrScript_lex_init_ALREADY_DEFINED
#else
#define yylex_init InstrScript_lex_init
#endif

#ifdef yylex_init_extra
#define InstrScript_lex_init_extra_ALREADY_DEFINED
#else
#define yylex_init_extra InstrScript_lex_init_extra
#endif

#ifdef yylex_destroy
#define InstrScript_lex_destroy_ALREADY_DEFINED
#else
#define yylex_destroy InstrScript_lex_destroy
#endif

#ifdef yyget_debug
#define InstrScript_get_debug_ALREADY_DEFINED
#else
#define yyget_debug InstrScript_get_debug
#endif

#ifdef yyset_debug
#define InstrScript_set_debug_ALREADY_DEFINED
#else
#define yyset_debug InstrScript_set_debug
#endif

#ifdef yyget_extra
#define InstrScript_get_extra_ALREADY_DEFINED
#else
#define yyget_extra InstrScript_get_extra
#endif

#ifdef yyset_extra
#define InstrScript_set_extra_ALREADY_DEFINED
#else
#define yyset_extra InstrScript_set_extra
#endif

#ifdef yyget_in
#define InstrScript_get_in_ALREADY_DEFINED
#else
#define yyget_in InstrScript_get_in
#endif

#ifdef yyset_in
#define InstrScript_set_in_ALREADY_DEFINED
#else
#define yyset_in InstrScript_set_in
#endif

#ifdef yyget_out
#define InstrScript_get_out_ALREADY_DEFINED
#else
#define yyget_out InstrScript_get_out
#endif

#ifdef yyset_out
#define InstrScript_set_out_ALREADY_DEFINED
#else
#define yyset_out InstrScript_set_out
#endif

#ifdef yyget_leng
#define InstrScript_get_leng_ALREADY_DEFINED
#else
#define yyget_leng InstrScript_get_leng
#endif

#ifdef yyget_text
#define InstrScript_get_text_ALREADY_DEFINED
#else
#define yyget_text InstrScript_get_text
#endif

#ifdef yyget_lineno
#define InstrScript_get_lineno_ALREADY_DEFINED
#else
#define yyget_lineno InstrScript_get_lineno
#endif

#ifdef yyset_lineno
#define InstrScript_set_lineno_ALREADY_DEFINED
#else
#define yyset_lineno InstrScript_set_lineno
#endif

#ifdef yyget_column
#define InstrScript_get_column_ALREADY_DEFINED
#else
#define yyget_column InstrScript_get_column
#endif

#ifdef yyset_column
#define InstrScript_set_column_ALREADY_DEFINED
#else
#define yyset_column InstrScript_set_column
#endif

#ifdef yywrap
#define InstrScript_wrap_ALREADY_DEFINED
#else
#define yywrap InstrScript_wrap
#endif

#ifdef yyget_lval
#define InstrScript_get_lval_ALREADY_DEFINED
#else
#define yyget_lval InstrScript_get_lval
#endif

#ifdef yyset_lval
#define InstrScript_set_lval_ALREADY_DEFINED
#else
#define yyset_lval InstrScript_set_lval
#endif

#ifdef yyget_lloc
#define InstrScript_get_lloc_ALREADY_DEFINED
#else
#define yyget_lloc InstrScript_get_lloc
#endif

#ifdef yyset_lloc
#define InstrScript_set_lloc_ALREADY_DEFINED
#else
#define yyset_lloc InstrScript_set_lloc
#endif

#ifdef yyalloc
#define InstrScript_alloc_ALREADY_DEFINED
#else
#define yyalloc InstrScript_alloc
#endif

#ifdef yyrealloc
#define InstrScript_realloc_ALREADY_DEFINED
#else
#define yyrealloc InstrScript_realloc
#endif

#ifdef yyfree
#define InstrScript_free_ALREADY_DEFINED
#else
#define yyfree InstrScript_free
#endif

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

/* end standard C headers. */

/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types. 
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <inttypes.h>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t; 
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN               (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN              (-32767-1)
#endif
#ifndef INT32_MIN
#define INT32_MIN              (-2147483647-1)
#endif
#ifndef INT8_MAX
#define INT8_MAX               (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX              (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX              (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX              (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX             (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX             (4294967295U)
#endif

#ifndef SIZE_MAX
#define SIZE_MAX               (~(size_t)0)
#endif

#endif /* ! C99 */

#endif /* ! FLEXINT_H */

/* begin standard C++ headers. */

/* TODO: this is always defined, so inline it */
#define yyconst const

#if defined(__GNUC__) && __GNUC__ >= 3
#define yynoreturn __attribute__((__noreturn__))
#else
#define yynoreturn
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an
 *   integer in range [0..255] for use as an array index.
 */
#define YY_SC_TO_UI(c) ((YY_CHAR) (c))

/* An opaque pointer. */
#ifndef YY_TYPEDEF_YY_SCANNER_T
#define YY_TYPEDEF_YY_SCANNER_T
typedef void* yyscan_t;
#endif

/* For convenience, these vars (plus the bison vars far below)
   are macros in the reentrant scanner. */
#define yyin yyg->yyin_r
#define yyout yyg->yyout_r
#define yyextra yyg->yyextra_r
#define yyleng yyg->yyleng_r
#define yytext yyg->yytext_r
#define yylineno (YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno)
#define yycolumn (YY_CURRENT_BUFFER_LVALUE->yy_bs_column)
#define yy_flex_debug yyg->yy_flex_debug_r

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN yyg->yy_start = 1 + 2 *
/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START ((yyg->yy_start - 1) / 2)
#define YYSTATE YY_START
/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)
/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yyrestart( yyin , yyscanner )
#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k.
 * Moreover, YY_BUF_SIZE is 2*YY_READ_BUF_SIZE in the general case.
 * Ditto for the __ia64__ case accordingly.
 */
#define YY_BUF_SIZE 32768
#else
#define YY_BUF_SIZE 16384
#endif /* __ia64__ */
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2
    
    #define YY_LESS_LINENO(n)
    #define YY_LINENO_REWIND_TO(ptr)
    
/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		*yy_cp = yyg->yy_hold_char; \
		YY_RESTORE_YY_MORE_OFFSET \
		yyg->yy_c_buf_p = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up yytext again */ \
		} \
	while ( 0 )
#define unput(c) yyunput( c, yyg->yytext_ptr , yyscanner )

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
	{
	FILE *yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	int yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	int yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

    int yy_bs_lineno; /**< The line count. */
    int yy_bs_column; /**< The column count. */

	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yyrestart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2

	};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER ( yyg->yy_buffer_stack \
                          ? yyg->yy_buffer_stack[yyg->yy_buffer_stack_top] \
                          : NULL)
/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE yyg->yy_buffer_stack[yyg->yy_buffer_stack_top]

void yyrestart ( FILE *input_file , yyscan_t yyscanner );
void yy_switch_to_buffer ( YY_BUFFER_STATE new_buffer , yyscan_t yyscanner );
YY_BUFFER_STATE yy_create_buffer ( FILE *file, int size , yyscan_t yyscanner );
void yy_delete_buffer ( YY_BUFFER_STATE b , yyscan_t yyscanner );
void yy_flush_buffer ( YY_BUFFER_STATE b , yyscan_t yyscanner );
void yypush_buffer_state ( YY_BUFFER_STATE new_buffer , yyscan_t yyscanner );
void yypop_buffer_state ( yyscan_t yyscanner );

static void yyensure_buffer_stack ( yyscan_t yyscanner );
static void yy_load_buffer_state ( yyscan_t yyscanner );
static void yy_init_buffer ( YY_BUFFER_STATE b, FILE *file , yyscan_t yyscanner );
#define YY_FLUSH_BUFFER yy_flush_buffer( YY_CURRENT_BUFFER , yyscanner)

YY_BUFFER_STATE yy_scan_buffer ( char *base, yy_size_t size , yyscan_t yyscanner );
YY_BUFFER_STATE yy_scan_string ( const char *yy_str , yyscan_t yyscanner );
YY_BUFFER_STATE yy_scan_bytes ( const char *bytes, int len , yyscan_t yyscanner );

void *yyalloc ( yy_size_t , yyscan_t yyscanner );
void *yyrealloc ( void *, yy_size_t , yyscan_t yyscanner );
void yyfree ( void * , yyscan_t yyscanner );

#define yy_new_buffer yy_create_buffer
#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){ \
        yyensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer( yyin, YY_BUF_SIZE , yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive; \
	}
#define yy_set_bol(at_bol) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){\
        yyensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer( yyin, YY_BUF_SIZE , yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol; \
	}
#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* Begin user sect3 */

#define InstrScript_wrap(yyscanner) (/*CONSTCOND*/1)
#define YY_SKIP_YYWRAP
typedef flex_uint8_t YY_CHAR;

typedef int yy_state_type;

#define yytext_ptr yytext_r

static yy_state_type yy_get_previous_state ( yyscan_t yyscanner );
static yy_state_type yy_try_NUL_trans ( yy_state_type current_state  , yyscan_t yyscanner);
static int yy_get_next_buffer ( yyscan_t yyscanner );
static void yynoreturn yy_fatal_error ( const char* msg , yyscan_t yyscanner );

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	yyg->yytext_ptr = yy_bp; \
	yyleng = (int) (yy_cp - yy_bp); \
	yyg->yy_hold_char = *yy_cp; \
	*yy_cp = '\0'; \
	yyg->yy_c_buf_p = yy_cp;
#define YY_NUM_RULES 63
#define YY_END_OF_BUFFER 64
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
	{
	flex_int32_t yy_verify;
	flex_int32_t yy_nxt;
	};
static const flex_int16_t yy_accept[563] =
    {   0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,   21,   21,    0,    0,    0,    0,   64,   62,
       61,   59,   55,   62,   62,   55,   62,    2,   62,   55,
       55,   57,   57,   57,   57,   57,   57,   57,   57,   57,
       57,   57,   57,   57,   57,   57,   57,   57,   57,   57,
       62,   57,   63,   63,    8,    7,    7,    7,    7,   63,
       11,   10,   10,   10,   10,   16,   16,   16,   16,   17,
       17,   17,   17,   22,   21,   21,   22,   22,   22,   22,
       63,   18,   63,   63,   63,   63,   19,   61,   59,    0,
        1,   56,   46,    0,    0,    0,    0,    2,    4,    0,

        4,    4,   58,   44,   45,   57,   57,   57,   57,   57,
       57,   57,   57,   57,   57,   57,   29,   57,   57,   57,
       57,   57,   23,   31,   57,   57,   57,   57,   57,   57,
       43,   57,    0,   60,   57,    0,    8,    7,    7,    7,
        7,    0,   11,   10,   10,   10,   10,   16,   16,   16,
       16,   17,   17,   17,   17,   22,   21,   21,   22,   22,
       22,   22,    0,   18,    0,    0,    0,    0,   19,   56,
        0,    0,    0,    3,    4,    4,   57,   57,   57,   34,
       57,   57,   57,   57,   57,   24,   57,   57,   57,   51,
       36,   57,   57,   57,   57,   39,   57,   57,   57,   57,

        7,    7,    7,   10,   10,   10,   16,   16,   16,   17,
       17,   17,   22,   22,   22,   22,    0,    0,    0,    0,
        0,   30,    5,    0,    5,    5,   57,   57,   57,   53,
       41,   57,   57,   57,   37,   57,   26,   57,   25,   40,
       57,   57,   57,   57,   57,   57,   57,    7,    7,    7,
       10,   10,   10,   16,   16,   16,   17,   17,   17,   22,
       22,   22,   22,    0,    0,    0,   33,   35,    5,    5,
       57,   57,   57,   47,   57,   57,   57,   50,   49,   57,
       57,   57,   57,   28,   57,    7,    7,    7,   10,   10,
       10,   16,   16,   16,   17,   17,   17,   22,   22,   22,

       22,    0,    0,    0,   57,   57,   57,   57,   57,   57,
       57,   57,   42,   57,   57,    7,    7,    7,   10,   10,
       10,   16,   16,   16,   17,   17,   17,   22,   22,   22,
       22,    0,    0,    0,   57,   57,   57,   57,   27,   57,
       57,   32,   57,   57,    7,    7,    7,   10,   10,   10,
       16,   16,   16,   17,   17,   17,   22,   22,   22,   22,
        0,    0,    0,   57,   57,   57,   57,   52,   57,   57,
       57,    7,    7,    7,   10,   10,   10,   16,   16,   16,
       17,   17,   17,   22,   22,   22,   22,    0,    0,    0,
       57,   57,   57,   57,   57,   57,   57,    7,    7,    7,

       10,   10,   10,   16,   16,   16,   17,   17,   17,   22,
       22,   22,   22,    0,    0,    0,   57,   57,   57,   38,
       48,   57,   57,    7,    7,    7,   10,   10,   10,   16,
       16,   16,   17,   17,   17,   22,   22,   22,   22,    0,
        0,    0,   57,   57,   57,   57,   57,    7,    7,    7,
       10,   10,   10,   16,   16,   16,   17,   17,   17,   22,
       22,   22,   22,    0,    0,    0,   57,   57,    0,   12,
       57,   54,   20,    7,    7,    7,   10,   10,   10,   16,
       16,   16,   17,   17,   17,   20,   22,   22,   22,   12,
       22,    0,    0,    0,   57,   57,   57,    7,    7,    7,

       10,   10,   10,   16,   16,   16,   17,   17,   17,   22,
       22,   22,   22,    0,    0,    0,   57,    0,    6,   57,
        7,    7,   10,   10,   16,   16,   17,   17,   22,   22,
        6,   13,   22,   22,    0,    0,   57,   57,    7,    7,
       10,   10,   16,   16,   17,   17,   22,   22,    0,    0,
        0,    9,    0,   14,   22,    9,   22,   14,   22,   15,
       22,    0
    } ;

static const YY_CHAR yy_ec[256] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    2,    3,
        1,    1,    4,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    2,    5,    6,    5,    7,    8,    5,    1,    9,
       10,    5,    5,    5,   11,   12,    5,   13,   13,   13,
       13,   13,   13,   13,   13,   13,   13,   14,    1,   15,
       16,   17,   18,   19,   20,   21,   22,   23,   24,   25,
       20,   26,   27,   20,   20,   20,   20,   28,   29,   20,
       20,   30,   31,   32,   33,   20,   20,   20,   20,   20,
        5,    1,    5,    1,   34,    1,   35,   20,   36,   37,

       38,   39,   20,   40,   41,   20,   42,   43,   44,   45,
       46,   47,   20,   48,   49,   50,   51,   20,   52,   20,
       53,   54,   55,    1,   56,   57,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,

        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1
    } ;

static const YY_CHAR yy_meta[58] =
    {   0,
        1,    1,    2,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    3,    1,    1,    1,    1,    1,    1,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    3,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    1,    1,    1
    } ;

static const flex_int16_t yy_base[579] =
    {   0,
        0,    0, 1047, 1046,   57,    0,  114,    0,  171,    0,
      228,    0,  284,  288,  292,  296,  301,  314, 1069, 1072,
      291, 1065, 1072, 1061,    0, 1049,  254,  336, 1049, 1048,
     1047,    0, 1038, 1037, 1029, 1014,  261, 1020,  285, 1006,
      265, 1010, 1009,  287,  291,  274,  299,  300, 1008, 1013,
      996, 1023, 1072,  295, 1072,    0, 1026, 1025, 1017,  340,
     1072,    0, 1023, 1022, 1014,    0, 1020, 1019, 1011,    0,
     1017, 1016, 1008,    0,  352,  356, 1010, 1013, 1012, 1004,
      341, 1072, 1010, 1009, 1001,  359, 1072,  361, 1028, 1024,
     1072,    0, 1072,  984,  982,  979, 1013,    0, 1072,  971,

      367,  989, 1072, 1072, 1072,    0,  992,  990,  997,  983,
      321,  974,  982,  968,  979,  970,    0,  973,  977,  975,
      961,  963,    0,    0,  959,  965,  964,  961,  962,  959,
        0,  962,  946, 1072,  978,  364, 1072,    0,  969,  967,
      974,  365, 1072,    0,  966,  964,  971,    0,  963,  961,
      968,    0,  960,  958,  965,    0,  379,  387,  965,  956,
      954,  961,  369, 1072,  953,  951,  958,  382, 1072,    0,
      944,  930,  967,  406,  373,  375,  954,  943,  942,    0,
      932,  936,  291,  930,  934,    0,  935,  920,  934,    0,
      930,  922,  930,  912,  926,    0,  925,  926,  918,  926,

      935,  924,  923,  932,  921,  920,  929,  918,  917,  926,
      915,  914,  913,  922,  911,  910,  919,  908,  907,  928,
      927, 1072, 1072,  884,  437,  902,  904,  913,  912,    0,
        0,  883,  884,  896,    0,  880,    0,  886,    0,    0,
      888,  880,  891,  889,  884,  885,  889,  889,  898,  897,
      886,  895,  894,  883,  892,  891,  880,  889,  888,  876,
      876,  885,  884,  873,  882,  881, 1072, 1072,  374,  391,
      868,  872,  871,    0,  853,  850,  856,    0,    0,  856,
      846,  844,  845,    0,  861,  857,  861,  860,  854,  858,
      857,  851,  855,  854,  848,  852,  851,  848,  842,  843,

      839,  830,  829,  798,  804,  797,  801,  780,  784,  775,
      774,  781,    0,  772,  793,  794,  787,  791,  791,  784,
      788,  788,  781,  785,  785,  778,  782,  780,  781,  774,
      778,  778,  771,  775,  768,  773,  771,  751,    0,  748,
      741,    0,  731,  736,  740,  745,  743,  737,  742,  740,
      734,  739,  737,  731,  736,  734,  723,  727,  732,  730,
      724,  729,  727,  722,  722,  714,  709,    0,  705,  704,
      722,  715,  715,  707,  712,  712,  704,  709,  709,  695,
      691,  686,  678,  689,  682,  682,  674,  679,  679,  671,
      681,  671,  675,  653,  664,  645,  669,  674,  664,  668,

      671,  660,  664,  667,  657,  661,  664,  654,  658,  655,
      660,  650,  654,  657,  647,  635,  629,  628,  629,    0,
        0,  615,  629,  624,  623,  624,  621,  620,  621,  618,
      617,  618,  615,  614,  615,  616,  611,  610,  611,  608,
      607,  608,  600,  602,  396,  593,  605,  596,  598,  404,
      594,  596,  419,  592,  594,  427,  590,  586,  432,  581,
      567,  569,  435,  565,  567,  450,  568,  566,  406, 1072,
      565,    0,    0,  565,  563,  562,  562,  560,  559,  559,
      557,  556,  556,  554,  553,    0,  553,  551,  445,  488,
      550,  550,  548,  544,  542,  447,  541,  540,  458,  513,

      511,  462,  510,  509,  463,  508,  462,  466,  453,  441,
      474,  545,  433,  422,  476,  416,  411,  478, 1072,  403,
      405,  399,  397,  388,  386,  370,  369,  354,  349,  506,
        0,    0,  602,  336,  324,  281,  507,  508,  509,  510,
      511,  512,  521,  522,  523,  524,  525,  526,  527,  539,
      556, 1072,  564, 1072,  565,    0,  566,  659,  716,    0,
      773, 1072,  830,  834,  838,  297,  840,  844,  846,  848,
      850,  852,  856,  858,  862,  866,  870,  874
    } ;

static const flex_int16_t yy_def[579] =
    {   0,
      562,    1,    1,    1,  562,    5,  562,    7,  562,    9,
      562,   11,  563,  563,  564,  564,  564,  564,  562,  562,
      562,  562,  562,  565,  566,  562,  562,  562,  562,  562,
      562,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      568,  567,  562,  562,  562,  569,  569,  569,  569,  562,
      562,  570,  570,  570,  570,  571,  571,  571,  571,  572,
      572,  572,  572,  573,  573,  562,  573,  573,  573,  573,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  565,
      562,  574,  562,  562,  562,  562,  562,   28,  562,  562,

      562,  101,  562,  562,  562,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  568,  562,  567,  562,  562,  569,  569,  569,
      569,  562,  562,  570,  570,  570,  570,  571,  571,  571,
      571,  572,  572,  572,  572,  573,  573,  562,  573,  573,
      573,  573,  562,  562,  562,  562,  562,  562,  562,  574,
      562,  562,  562,  562,  562,  562,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,

      569,  569,  569,  570,  570,  570,  571,  571,  571,  572,
      572,  572,  573,  573,  573,  573,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  225,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  567,  567,  569,  569,  569,
      570,  570,  570,  571,  571,  571,  572,  572,  572,  573,
      573,  573,  573,  562,  562,  562,  562,  562,  562,  562,
      567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  569,  569,  569,  570,  570,
      570,  571,  571,  571,  572,  572,  572,  573,  573,  573,

      573,  562,  562,  562,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  567,  569,  569,  569,  570,  570,
      570,  571,  571,  571,  572,  572,  572,  573,  573,  573,
      573,  562,  562,  562,  567,  567,  567,  567,  567,  567,
      567,  567,  567,  567,  569,  569,  569,  570,  570,  570,
      571,  571,  571,  572,  572,  572,  573,  573,  573,  573,
      562,  562,  562,  567,  567,  567,  567,  567,  567,  567,
      567,  569,  569,  569,  570,  570,  570,  571,  571,  571,
      572,  572,  572,  573,  573,  573,  573,  562,  562,  562,
      567,  567,  567,  567,  567,  567,  567,  569,  569,  569,

      570,  570,  570,  571,  571,  571,  572,  572,  572,  573,
      573,  573,  573,  562,  562,  562,  567,  567,  567,  567,
      567,  567,  567,  569,  569,  569,  570,  570,  570,  571,
      571,  571,  572,  572,  572,  573,  573,  573,  573,  562,
      562,  562,  567,  567,  567,  567,  567,  569,  569,  569,
      570,  570,  570,  571,  571,  571,  572,  572,  572,  573,
      573,  573,  573,  562,  562,  562,  567,  567,  562,  562,
      567,  567,  567,  569,  569,  569,  570,  570,  570,  571,
      571,  571,  572,  572,  572,  573,  573,  573,  573,  575,
      573,  562,  562,  562,  567,  567,  567,  569,  569,  569,

      570,  570,  570,  571,  571,  571,  572,  572,  572,  573,
      573,  576,  573,  562,  562,  562,  567,  562,  562,  567,
      569,  569,  570,  570,  571,  571,  572,  572,  573,  573,
      573,  573,  576,  573,  562,  562,  567,  567,  569,  569,
      570,  570,  571,  571,  572,  572,  573,  573,  562,  562,
      562,  562,  562,  562,  573,  573,  573,  577,  578,  573,
      578,    0,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562
    } ;

static const flex_int16_t yy_nxt[1130] =
    {   0,
       20,   21,   22,   21,   23,   24,   25,   25,   23,   23,
       26,   27,   28,   29,   30,   23,   31,   25,   25,   32,
       32,   32,   32,   32,   32,   32,   32,   32,   32,   33,
       34,   32,   35,   20,   36,   37,   38,   39,   40,   32,
       41,   32,   42,   43,   44,   45,   46,   47,   48,   49,
       32,   50,   32,   32,   51,   20,   25,   53,   54,   53,
       53,   53,   53,   53,   53,   53,   55,   53,   53,   53,
       53,   53,   53,   53,   53,   53,   56,   56,   56,   56,
       56,   56,   56,   56,   56,   56,   57,   58,   56,   59,
       53,   56,   56,   56,   56,   56,   56,   56,   56,   56,

       56,   56,   56,   56,   56,   56,   56,   56,   56,   56,
       56,   53,   53,   53,   53,   60,   53,   53,   53,   53,
       53,   53,   53,   61,   53,   53,   53,   53,   53,   53,
       53,   53,   53,   62,   62,   62,   62,   62,   62,   62,
       62,   62,   62,   63,   64,   62,   65,   53,   62,   62,
       62,   62,   62,   62,   62,   62,   62,   62,   62,   62,
       62,   62,   62,   62,   62,   62,   62,   62,   53,   53,
       53,   53,   53,   53,   53,   53,   53,   53,   53,   53,
       53,   53,   53,   53,   53,   53,   53,   53,   53,   53,
       66,   66,   66,   66,   66,   66,   66,   66,   66,   66,

       67,   68,   66,   69,   53,   66,   66,   66,   66,   66,
       66,   66,   66,   66,   66,   66,   66,   66,   66,   66,
       66,   66,   66,   66,   66,   53,   53,   53,   53,   53,
       53,   53,   53,   53,   53,   53,   53,   53,   53,   53,
       53,   53,   53,   53,   53,   53,   53,   70,   70,   70,
       70,   70,   70,   70,   70,   70,   70,   71,   72,   70,
       73,   53,   70,   70,   70,   70,   70,   70,   70,   70,
       70,   70,   70,   70,   70,   70,   70,   70,   70,   70,
       70,   70,   53,   53,   53,   75,   76,   75,   94,   75,
       76,   75,   88,   81,   88,  111,  136,   81,   95,   96,

       92,   82,   86,  117,  137,   82,  112,   77,  125,  118,
       87,   77,  550,   78,   79,   86,   80,   78,   79,  126,
       80,   83,   84,   87,   85,   83,   84,  114,   85,  115,
       83,   84,  121,   85,  122,  123,  127,  129,  124,  232,
      233,  142,  163,   83,   84,  128,   85,   97,   98,  143,
      164,  549,  130,  157,  158,  157,   99,  158,  158,  158,
      168,  100,   88,  181,   88,  136,  142,  548,  169,  182,
      163,  101,  102,  137,  143,  101,  547,  101,  164,  101,
      157,  158,  157,  168,   99,  546,  101,   99,  158,  158,
      158,  169,  100,   99,  223,   99,  545,  469,  100,  224,

      100,  544,  175,  176,  470,  469,  175,  469,  175,  175,
      175,  223,  470,  543,  470,   99,  224,  175,  174,  542,
      469,   99,  223,   99,  541,  269,  223,  470,  469,  471,
      540,  224,  539,  469,  538,  470,  489,  476,  537,  223,
      470,  225,  226,  490,  536,  225,  489,  225,  518,  225,
      535,  469,  479,  490,  223,  519,  225,  223,  470,  518,
      482,  534,  224,  518,  518,  485,  519,  518,  491,  529,
      519,  519,  269,  270,  519,  530,  269,  518,  269,  518,
      269,  528,  531,  494,  519,  223,  519,  269,  156,  156,
      527,  156,  156,  156,  156,  156,  156,  156,  156,  156,

      156,  156,  156,  156,  156,  156,  156,  530,  551,  553,
      551,  553,  551,  553,  531,  552,  554,  552,  554,  552,
      554,  156,  551,  553,  551,  553,  555,  557,  551,  552,
      554,  552,  554,  556,  558,  552,  526,  525,  524,  523,
      553,  522,  156,  156,  156,  156,  156,  554,  156,  156,
      156,  156,  156,  156,  532,  156,  156,  551,  156,  156,
      156,  156,  156,  156,  552,  553,  555,  557,  521,  520,
      517,  516,  554,  556,  558,  515,  514,  513,  511,  510,
      509,  508,  507,  506,  505,  504,  503,  502,  501,  500,
      499,  498,  497,  496,  495,  493,  492,  488,  487,  156,

      156,  156,  156,  156,  486,  156,  156,  156,  156,  156,
      156,  532,  156,  156,  484,  156,  156,  156,  156,  156,
      156,  483,  481,  480,  478,  477,  475,  474,  473,  472,
      468,  467,  466,  465,  464,  463,  462,  461,  460,  459,
      458,  457,  456,  455,  454,  453,  452,  451,  450,  449,
      448,  447,  446,  445,  444,  443,  156,  156,  156,  156,
      156,  442,  156,  156,  156,  156,  156,  156,  156,  156,
      156,  156,  156,  156,  156,  156,  156,  156,  441,  440,
      439,  438,  437,  436,  435,  434,  433,  432,  431,  430,
      429,  428,  156,  427,  426,  425,  424,  423,  422,  421,

      420,  419,  418,  417,  416,  415,  414,  413,  412,  411,
      410,  409,  408,  156,  156,  156,  156,  156,  407,  156,
      156,  156,  156,  156,  156,  560,  156,  156,  406,  156,
      156,  156,  156,  156,  156,  405,  404,  403,  402,  401,
      400,  399,  398,  397,  396,  395,  394,  393,  392,  391,
      390,  389,  388,  387,  386,  385,  384,  383,  382,  381,
      380,  379,  378,  377,  376,  375,  374,  373,  372,  371,
      156,  156,  156,  156,  156,  370,  156,  156,  156,  156,
      156,  156,  560,  156,  156,  369,  156,  156,  156,  156,
      156,  156,  368,  367,  366,  365,  364,  363,  362,  361,

      360,  359,  358,  357,  356,  355,  354,  353,  352,  351,
      350,  349,  348,  347,  346,  345,  344,  343,  342,  341,
      340,  339,  338,  337,  336,  335,  334,  156,  156,  156,
       74,   74,   74,   74,   53,   53,   53,   53,   90,   90,
       90,   90,  106,  106,  133,  133,  133,  133,  138,  138,
      144,  144,  148,  148,  152,  152,  156,  333,  156,  156,
      170,  170,  512,  332,  512,  512,  533,  331,  533,  533,
      559,  330,  559,  559,  561,  329,  561,  561,  328,  327,
      326,  325,  324,  323,  322,  321,  320,  319,  318,  317,
      316,  315,  314,  313,  312,  311,  310,  309,  308,  307,

      306,  305,  304,  303,  302,  301,  300,  299,  298,  297,
      296,  295,  294,  293,  292,  291,  290,  289,  288,  287,
      286,  285,  284,  283,  282,  281,  280,  279,  278,  277,
      276,  275,  274,  273,  272,  271,  225,  223,  268,  267,
      266,  265,  264,  263,  262,  261,  260,  259,  258,  257,
      256,  255,  254,  253,  252,  251,  250,  249,  248,  247,
      246,  245,  244,  243,  242,  241,  240,  239,  238,  237,
      236,  235,  234,  231,  230,  229,  228,  227,  222,  221,
      220,  219,  218,  217,  216,  215,  214,  213,  212,  211,
      210,  209,  208,  207,  206,  205,  204,  203,  202,  201,

      200,  134,  199,  198,  197,  196,  195,  194,  193,  192,
      191,  190,  189,  188,  187,  186,  185,  184,  183,  180,
      179,  178,  177,  101,   99,  174,  173,  172,  171,   91,
       89,  167,  166,  165,  162,  161,  160,  159,  155,  154,
      153,  151,  150,  149,  147,  146,  145,  141,  140,  139,
      135,  134,  132,  131,  120,  119,  116,  113,  110,  109,
      108,  107,  105,  104,  103,   93,   91,   89,  562,   52,
       52,   19,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,

      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562
    } ;

static const flex_int16_t yy_chk[1130] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,

        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    7,    7,    7,    7,    7,    7,
        7,    7,    7,    7,    7,    7,    7,    7,    7,    7,
        7,    7,    7,    7,    7,    7,    7,    7,    7,    7,
        7,    7,    7,    7,    7,    7,    7,    7,    7,    7,
        7,    7,    7,    7,    7,    7,    7,    7,    7,    7,
        7,    7,    7,    7,    7,    7,    7,    7,    7,    7,
        7,    9,    9,    9,    9,    9,    9,    9,    9,    9,
        9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
        9,    9,    9,    9,    9,    9,    9,    9,    9,    9,

        9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
        9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
        9,    9,    9,    9,    9,    9,    9,    9,   11,   11,
       11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
       11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
       11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
       11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
       11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
       11,   11,   11,   11,   11,   13,   13,   13,   27,   14,
       14,   14,   21,   15,   21,   37,   54,   16,   27,   27,

      566,   15,   17,   41,   54,   16,   37,   13,   46,   41,
       17,   14,  536,   13,   13,   18,   13,   14,   14,   46,
       14,   15,   15,   18,   15,   16,   16,   39,   16,   39,
       17,   17,   44,   17,   44,   45,   47,   48,   45,  183,
      183,   60,   81,   18,   18,   47,   18,   28,   28,   60,
       81,  535,   48,   75,   75,   75,   28,   76,   76,   76,
       86,   28,   88,  111,   88,  136,  142,  534,   86,  111,
      163,   28,   28,  136,  142,   28,  529,   28,  163,   28,
      157,  157,  157,  168,   28,  528,   28,  101,  158,  158,
      158,  168,  101,  175,  269,  176,  527,  445,  175,  269,

      176,  526,  101,  101,  445,  450,  101,  469,  101,  176,
      101,  270,  450,  525,  469,  101,  270,  101,  174,  524,
      453,  175,  269,  176,  523,  270,  174,  453,  456,  445,
      522,  174,  521,  459,  520,  456,  463,  450,  517,  270,
      459,  174,  174,  463,  516,  174,  489,  174,  496,  174,
      514,  466,  453,  489,  174,  496,  174,  225,  466,  499,
      456,  513,  225,  502,  505,  459,  499,  508,  463,  510,
      502,  505,  225,  225,  508,  511,  225,  515,  225,  518,
      225,  509,  511,  466,  515,  225,  518,  225,  490,  490,
      507,  490,  490,  490,  490,  490,  490,  490,  490,  490,

      490,  490,  490,  490,  490,  490,  490,  530,  537,  538,
      539,  540,  541,  542,  530,  537,  538,  539,  540,  541,
      542,  490,  543,  544,  545,  546,  547,  548,  549,  543,
      544,  545,  546,  547,  548,  549,  506,  504,  503,  501,
      550,  500,  490,  490,  490,  512,  512,  550,  512,  512,
      512,  512,  512,  512,  512,  512,  512,  551,  512,  512,
      512,  512,  512,  512,  551,  553,  555,  557,  498,  497,
      495,  494,  553,  555,  557,  493,  492,  491,  488,  487,
      485,  484,  483,  482,  481,  480,  479,  478,  477,  476,
      475,  474,  471,  468,  467,  465,  464,  462,  461,  512,

      512,  512,  533,  533,  460,  533,  533,  533,  533,  533,
      533,  533,  533,  533,  458,  533,  533,  533,  533,  533,
      533,  457,  455,  454,  452,  451,  449,  448,  447,  446,
      444,  443,  442,  441,  440,  439,  438,  437,  436,  435,
      434,  433,  432,  431,  430,  429,  428,  427,  426,  425,
      424,  423,  422,  419,  418,  417,  533,  533,  533,  558,
      558,  416,  558,  558,  558,  558,  558,  558,  558,  558,
      558,  558,  558,  558,  558,  558,  558,  558,  415,  414,
      413,  412,  411,  410,  409,  408,  407,  406,  405,  404,
      403,  402,  558,  401,  400,  399,  398,  397,  396,  395,

      394,  393,  392,  391,  390,  389,  388,  387,  386,  385,
      384,  383,  382,  558,  558,  558,  559,  559,  381,  559,
      559,  559,  559,  559,  559,  559,  559,  559,  380,  559,
      559,  559,  559,  559,  559,  379,  378,  377,  376,  375,
      374,  373,  372,  371,  370,  369,  367,  366,  365,  364,
      363,  362,  361,  360,  359,  358,  357,  356,  355,  354,
      353,  352,  351,  350,  349,  348,  347,  346,  345,  344,
      559,  559,  559,  561,  561,  343,  561,  561,  561,  561,
      561,  561,  561,  561,  561,  341,  561,  561,  561,  561,
      561,  561,  340,  338,  337,  336,  335,  334,  333,  332,

      331,  330,  329,  328,  327,  326,  325,  324,  323,  322,
      321,  320,  319,  318,  317,  316,  315,  314,  312,  311,
      310,  309,  308,  307,  306,  305,  304,  561,  561,  561,
      563,  563,  563,  563,  564,  564,  564,  564,  565,  565,
      565,  565,  567,  567,  568,  568,  568,  568,  569,  569,
      570,  570,  571,  571,  572,  572,  573,  303,  573,  573,
      574,  574,  575,  302,  575,  575,  576,  301,  576,  576,
      577,  300,  577,  577,  578,  299,  578,  578,  298,  297,
      296,  295,  294,  293,  292,  291,  290,  289,  288,  287,
      286,  285,  283,  282,  281,  280,  277,  276,  275,  273,

      272,  271,  266,  265,  264,  263,  262,  261,  260,  259,
      258,  257,  256,  255,  254,  253,  252,  251,  250,  249,
      248,  247,  246,  245,  244,  243,  242,  241,  238,  236,
      234,  233,  232,  229,  228,  227,  226,  224,  221,  220,
      219,  218,  217,  216,  215,  214,  213,  212,  211,  210,
      209,  208,  207,  206,  205,  204,  203,  202,  201,  200,
      199,  198,  197,  195,  194,  193,  192,  191,  189,  188,
      187,  185,  184,  182,  181,  179,  178,  177,  173,  172,
      171,  167,  166,  165,  162,  161,  160,  159,  155,  154,
      153,  151,  150,  149,  147,  146,  145,  141,  140,  139,

      135,  133,  132,  130,  129,  128,  127,  126,  125,  122,
      121,  120,  119,  118,  116,  115,  114,  113,  112,  110,
      109,  108,  107,  102,  100,   97,   96,   95,   94,   90,
       89,   85,   84,   83,   80,   79,   78,   77,   73,   72,
       71,   69,   68,   67,   65,   64,   63,   59,   58,   57,
       52,   51,   50,   49,   43,   42,   40,   38,   36,   35,
       34,   33,   31,   30,   29,   26,   24,   22,   19,    4,
        3,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,

      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
      562,  562,  562,  562,  562,  562,  562,  562,  562
    } ;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
#line 1 "scanner.l"
/*
 * Copyright (c) 2014-2017 Christian Schoenebeck and Andreas Persson
 * Copyright (c) 2018-2025 Christian Schoenebeck
 *
 * http://www.linuxsampler.org
 *
 * This file is part of LinuxSampler and released under the same terms.
 * See README file for details.
 */
/* Token scanner for NKSP real-time instrument script language. */
#line 14 "scanner.l"

#include "parser_shared.h"
#include <math.h>
// reentrant scanner data context
#define YY_EXTRA_TYPE ParserContext*
// set line number each time a token is recognized
#define YY_USER_ACTION                  \
{                                       \
    yylloc->first_line = yylineno;      \
    yylloc->last_line  = yylineno;      \
    yylloc->first_column = yycolumn;    \
    yylloc->first_byte = yyextra->nbytes; \
    yylloc->length_bytes = (int) yyleng;  \
    yycolumn += yyleng;                 \
    yyextra->nbytes += (int) yyleng;    \
    yylloc->last_column = yycolumn - 1; \
    /*printf("lex: line '%s'\n", yytext);*/  \
}
// custom (f)lex input for reading from std::istream object
#define YY_INPUT(buf,result,max_size)   \
{                                       \
    char c = yyextra->is->get();        \
    if (yyextra->is->eof())             \
        result = YY_NULL;               \
    else {                              \
        buf[0] = c;                     \
        result = 1;                     \
    }                                   \
}

static void scanner_error(YYLTYPE* locp, LinuxSampler::ParserContext* context, const char* err) {
    context->addErr(locp->first_line, locp->last_line, locp->first_column,
                    locp->last_column, locp->first_byte, locp->length_bytes,
                    err);
}

static void scanner_warning(YYLTYPE* locp, LinuxSampler::ParserContext* context, const char* txt) {
    context->addWrn(locp->first_line, locp->last_line, locp->first_column,
                    locp->last_column, locp->first_byte, locp->length_bytes,
                    txt);
}

#define SCANNER_ERR(txt)  scanner_error(yylloc, yyextra, txt)
#define SCANNER_WRN(txt)  scanner_warning(yylloc, yyextra, txt)

// shut up warning that 'register' keyword is deprecated as of C++11
#if defined(__cplusplus) && __cplusplus >= 201103L
# define register
#endif

using namespace LinuxSampler;

static int countNewLineChars(const char* txt) {
    int n = 0;
    for (int i = 0; txt[i]; ++i)
        if (txt[i] == '\n') ++n;
    return n;
}

static int countCharsPastLastNewLine(const char* txt) {
    const int n = (int)strlen(txt);
    for (int i = n - 1; i >= 0; --i)
        if (txt[i] == '\n')
            return n - i - 1;
    return n;
}

#define processLocation() { \
    const int nl = countNewLineChars(yytext); \
    yylineno += nl; \
    if (nl) yycolumn = countCharsPastLastNewLine(yytext); \
}

// if compiled for debugging, throw an exception instead of exiting on fatal
// lexer errors (so the debugger may pause with the appropriate back trace)
#if DEBUG
# include <stdexcept>
# define YY_FATAL_ERROR(msg) throw std::runtime_error(msg)
#endif

#line 1182 "scanner.cpp"
/* use Flex's built-in support for line numbers
   (disabled, because it seems to be unreliable, so we are using our own
   tracking code in the respective scanner rules below) */
/*%option yylineno*/
/* generate a reentrant safe scanner */
/* avoid symbol collision with other (i.e. future) scanner symbols */
/* bison-bridge adds an argument yylval to yylex, and bison-locations adds an
   argument code yylloc for location tracking. */
/* yywrap() would be called at EOF, we don't need it */
/* enable functions yy_push_state(), yy_pop_state(), yy_top_state() */
/* inclusive scanner conditions */

/* exclusive scanner conditions */

#line 1197 "scanner.cpp"

#define INITIAL 0
#define PREPROC_BODY_USE 1
#define PREPROC_SET_COND 2
#define PREPROC_RESET_COND 3
#define PREPROC_IF 4
#define PREPROC_IF_NOT 5
#define PREPROC_BODY_EAT 6
#define PREPROC_PRE_BODY_USE 7
#define PREPROC_PRE_BODY_EAT 8

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
#include <unistd.h>
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

/* Holds the entire state of the reentrant scanner. */
struct yyguts_t
    {

    /* User-defined. Not touched by flex. */
    YY_EXTRA_TYPE yyextra_r;

    /* The rest are the same as the globals declared in the non-reentrant scanner. */
    FILE *yyin_r, *yyout_r;
    size_t yy_buffer_stack_top; /**< index of top of stack. */
    size_t yy_buffer_stack_max; /**< capacity of stack. */
    YY_BUFFER_STATE * yy_buffer_stack; /**< Stack as an array. */
    char yy_hold_char;
    int yy_n_chars;
    int yyleng_r;
    char *yy_c_buf_p;
    int yy_init;
    int yy_start;
    int yy_did_buffer_switch_on_eof;
    int yy_start_stack_ptr;
    int yy_start_stack_depth;
    int *yy_start_stack;
    yy_state_type yy_last_accepting_state;
    char* yy_last_accepting_cpos;

    int yylineno_r;
    int yy_flex_debug_r;

    char *yytext_r;
    int yy_more_flag;
    int yy_more_len;

    YYSTYPE * yylval_r;

    YYLTYPE * yylloc_r;

    }; /* end struct yyguts_t */

static int yy_init_globals ( yyscan_t yyscanner );

    /* This must go here because YYSTYPE and YYLTYPE are included
     * from bison output in section 1.*/
    #    define yylval yyg->yylval_r
    
    #    define yylloc yyg->yylloc_r
    
int yylex_init (yyscan_t* scanner);

int yylex_init_extra ( YY_EXTRA_TYPE user_defined, yyscan_t* scanner);

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int yylex_destroy ( yyscan_t yyscanner );

int yyget_debug ( yyscan_t yyscanner );

void yyset_debug ( int debug_flag , yyscan_t yyscanner );

YY_EXTRA_TYPE yyget_extra ( yyscan_t yyscanner );

void yyset_extra ( YY_EXTRA_TYPE user_defined , yyscan_t yyscanner );

FILE *yyget_in ( yyscan_t yyscanner );

void yyset_in  ( FILE * _in_str , yyscan_t yyscanner );

FILE *yyget_out ( yyscan_t yyscanner );

void yyset_out  ( FILE * _out_str , yyscan_t yyscanner );

			int yyget_leng ( yyscan_t yyscanner );

char *yyget_text ( yyscan_t yyscanner );

int yyget_lineno ( yyscan_t yyscanner );

void yyset_lineno ( int _line_number , yyscan_t yyscanner );

int yyget_column  ( yyscan_t yyscanner );

void yyset_column ( int _column_no , yyscan_t yyscanner );

YYSTYPE * yyget_lval ( yyscan_t yyscanner );

void yyset_lval ( YYSTYPE * yylval_param , yyscan_t yyscanner );

       YYLTYPE *yyget_lloc ( yyscan_t yyscanner );
    
        void yyset_lloc ( YYLTYPE * yylloc_param , yyscan_t yyscanner );
    
/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yywrap ( yyscan_t yyscanner );
#else
extern int yywrap ( yyscan_t yyscanner );
#endif
#endif

#ifndef YY_NO_UNPUT
    
    static void yyunput ( int c, char *buf_ptr  , yyscan_t yyscanner);
    
#endif

#ifndef yytext_ptr
static void yy_flex_strncpy ( char *, const char *, int , yyscan_t yyscanner);
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen ( const char * , yyscan_t yyscanner);
#endif

#ifndef YY_NO_INPUT
#ifdef __cplusplus
static int yyinput ( yyscan_t yyscanner );
#else
static int input ( yyscan_t yyscanner );
#endif

#endif

    static void yy_push_state ( int _new_state , yyscan_t yyscanner);
    
    static void yy_pop_state ( yyscan_t yyscanner );
    
    static int yy_top_state ( yyscan_t yyscanner );
    
/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k */
#define YY_READ_BUF_SIZE 16384
#else
#define YY_READ_BUF_SIZE 8192
#endif /* __ia64__ */
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO do { if (fwrite( yytext, (size_t) yyleng, 1, yyout )) {} } while (0)
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if ( YY_CURRENT_BUFFER_LVALUE->yy_is_interactive ) \
		{ \
		int c = '*'; \
		int n; \
		for ( n = 0; n < max_size && \
			     (c = getc( yyin )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( yyin ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else \
		{ \
		errno=0; \
		while ( (result = (int) fread(buf, 1, (yy_size_t) max_size, yyin)) == 0 && ferror(yyin)) \
			{ \
			if( errno != EINTR) \
				{ \
				YY_FATAL_ERROR( "input in flex scanner failed" ); \
				break; \
				} \
			errno=0; \
			clearerr(yyin); \
			} \
		}\
\

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error( msg , yyscanner)
#endif

/* end tables serialization structures and prototypes */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1

extern int yylex \
               (YYSTYPE * yylval_param, YYLTYPE * yylloc_param , yyscan_t yyscanner);

#define YY_DECL int yylex \
               (YYSTYPE * yylval_param, YYLTYPE * yylloc_param , yyscan_t yyscanner)
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK /*LINTED*/break;
#endif

#define YY_RULE_SETUP \
	YY_USER_ACTION

/** The main scanner function which does all the work.
 */
YY_DECL
{
	yy_state_type yy_current_state;
	char *yy_cp, *yy_bp;
	int yy_act;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    yylval = yylval_param;

    yylloc = yylloc_param;

	if ( !yyg->yy_init )
		{
		yyg->yy_init = 1;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! yyg->yy_start )
			yyg->yy_start = 1;	/* first start state */

		if ( ! yyin )
			yyin = stdin;

		if ( ! yyout )
			yyout = stdout;

		if ( ! YY_CURRENT_BUFFER ) {
			yyensure_buffer_stack (yyscanner);
			YY_CURRENT_BUFFER_LVALUE =
				yy_create_buffer( yyin, YY_BUF_SIZE , yyscanner);
		}

		yy_load_buffer_state( yyscanner );
		}

	{
#line 123 "scanner.l"


#line 1498 "scanner.cpp"

	while ( /*CONSTCOND*/1 )		/* loops until end-of-file is reached */
		{
		yy_cp = yyg->yy_c_buf_p;

		/* Support of yytext. */
		*yy_cp = yyg->yy_hold_char;

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

		yy_current_state = yyg->yy_start;
yy_match:
		do
			{
			YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)] ;
			if ( yy_accept[yy_current_state] )
				{
				yyg->yy_last_accepting_state = yy_current_state;
				yyg->yy_last_accepting_cpos = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 563 )
					yy_c = yy_meta[yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
			++yy_cp;
			}
		while ( yy_base[yy_current_state] != 1072 );

yy_find_action:
		yy_act = yy_accept[yy_current_state];
		if ( yy_act == 0 )
			{ /* have to back up */
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			yy_act = yy_accept[yy_current_state];
			}

		YY_DO_BEFORE_ACTION;

do_action:	/* This label is used only to access EOF actions. */

		switch ( yy_act )
	{ /* beginning of action switch */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = yyg->yy_hold_char;
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			goto yy_find_action;

case 1:
/* rule 1 can match eol */
YY_RULE_SETUP
#line 125 "scanner.l"
{
    char* s = strdup(yytext + 1);
    s[strlen(s) - 1] = '\0';
    yylval->sValue = s;
    yyextra->autoFreeAfterParse(s);
    return STRING;
}
	YY_BREAK
case 2:
YY_RULE_SETUP
#line 133 "scanner.l"
{
    if (sizeof(vmint) < 8)
        yylval->iValue = atoi(yytext);
    else
        yylval->iValue = atoll(yytext);
    return INTEGER;
}
	YY_BREAK
case 3:
YY_RULE_SETUP
#line 141 "scanner.l"
{
    yylval->fValue = strTo<vmfloat>(yytext);
    return REAL;
}
	YY_BREAK
case 4:
YY_RULE_SETUP
#line 146 "scanner.l"
{
    int pos = 0;

    // parse number portion
    vmint value = 0;
    for (; yytext[pos] >= '0' && yytext[pos] <= '9'; ++pos) {
        value *= 10;
        value += yytext[pos] - '0';
    }
    yylval->iUnitValue.iValue = value;

    // parse metric prefix portion
    for (int i = 0; i < 2; ++i, ++pos) {
        switch (yytext[pos]) {
            case 'k': yylval->iUnitValue.prefix[i] = VM_KILO;  continue;
            case 'h': yylval->iUnitValue.prefix[i] = VM_HECTO; continue;
            case 'c': yylval->iUnitValue.prefix[i] = VM_CENTI; continue;
            case 'm': yylval->iUnitValue.prefix[i] = VM_MILLI; continue;
            case 'u': yylval->iUnitValue.prefix[i] = VM_MICRO; continue;
            case 'd':
                if (yytext[pos+1] == 'a') {
                    yylval->iUnitValue.prefix[i] = VM_DECA;
                    ++pos;
                } else {
                    yylval->iUnitValue.prefix[i] = VM_DECI;
                }
                continue;
            default:
                yylval->iUnitValue.prefix[i] = VM_NO_PREFIX;
                goto parseIntStdUnit;
        }
    }

    parseIntStdUnit:

    // parse standard measurement unit
    switch (yytext[pos]) {
        case 's': yylval->iUnitValue.unit = VM_SECOND;  break;
        case 'H': yylval->iUnitValue.unit = VM_HERTZ;   break;
        case 'B': yylval->iUnitValue.unit = VM_BEL;     break;
        default:  yylval->iUnitValue.unit = VM_NO_UNIT; break;
    }

    return INTEGER_UNIT;
}
	YY_BREAK
case 5:
YY_RULE_SETUP
#line 192 "scanner.l"
{
    int pos = 0;

    // parse number portion
    for (; (yytext[pos] >= '0' && yytext[pos] <= '9') || yytext[pos] == '.'; ++pos) {
    }
    {
        const char tmp = yytext[pos];
        yytext[pos] = 0; // mark temporary end of string
        yylval->fUnitValue.fValue = strTo<vmfloat>(yytext);
        yytext[pos] = tmp; // restore
    }

    // parse metric prefix portion
    for (int i = 0; i < 2; ++i, ++pos) {
        switch (yytext[pos]) {
            case 'k': yylval->fUnitValue.prefix[i] = VM_KILO;  continue;
            case 'h': yylval->fUnitValue.prefix[i] = VM_HECTO; continue;
            case 'c': yylval->fUnitValue.prefix[i] = VM_CENTI; continue;
            case 'm': yylval->fUnitValue.prefix[i] = VM_MILLI; continue;
            case 'u': yylval->fUnitValue.prefix[i] = VM_MICRO; continue;
            case 'd':
                if (yytext[pos+1] == 'a') {
                    yylval->fUnitValue.prefix[i] = VM_DECA;
                    ++pos;
                } else {
                    yylval->fUnitValue.prefix[i] = VM_DECI;
                }
                continue;
            default:
                yylval->fUnitValue.prefix[i] = VM_NO_PREFIX;
                goto parseRealStdUnit;
        }
    }

    parseRealStdUnit:

    // parse standard measurement unit
    switch (yytext[pos]) {
        case 's': yylval->fUnitValue.unit = VM_SECOND;  break;
        case 'H': yylval->fUnitValue.unit = VM_HERTZ;   break;
        case 'B': yylval->fUnitValue.unit = VM_BEL;     break;
        default:  yylval->fUnitValue.unit = VM_NO_UNIT; break;
    }

    return REAL_UNIT;
}
	YY_BREAK
/* Preprocessor statement:  SET_CONDITION(name) */
case 6:
YY_RULE_SETUP
#line 243 "scanner.l"
{
    //printf("SET_CONDITION\n");
    yy_push_state(PREPROC_SET_COND, yyscanner);
}
	YY_BREAK
case 7:
YY_RULE_SETUP
#line 247 "scanner.l"
{
    //printf("preproc set condition '%s'\n", yytext);
    bool success = yyextra->setPreprocessorCondition(yytext);
    if (!success) {
        SCANNER_WRN((String("Preprocessor: Condition '") +
                    yytext + "' is already set.").c_str());
    }
}
	YY_BREAK
case 8:
YY_RULE_SETUP
#line 255 "scanner.l"
{
    //printf("End of PREPROC_SET_COND\n");
    yy_pop_state(yyscanner);
}
	YY_BREAK
/* Preprocessor statement:  RESET_CONDITION(name) */
case 9:
YY_RULE_SETUP
#line 263 "scanner.l"
{
    //printf("RESET_CONDITION\n");
    yy_push_state(PREPROC_RESET_COND, yyscanner);
}
	YY_BREAK
case 10:
YY_RULE_SETUP
#line 267 "scanner.l"
{
    //printf("preproc reset condition '%s'\n", yytext);
    bool success = yyextra->resetPreprocessorCondition(yytext);
    if (!success) {
        SCANNER_ERR((String("Preprocessor: could not reset condition '") +
                    yytext + "' (either not set or a built-in condition).").c_str());
    }
}
	YY_BREAK
case 11:
YY_RULE_SETUP
#line 275 "scanner.l"
{
    //printf("End of RESET_CONDITION\n");
    yy_pop_state(yyscanner);
}
	YY_BREAK
/* Preprocessor conditional statements:

    USE_CODE_IF(name)
      ...
    END_USE_CODE

    and:

    USE_CODE_IF_NOT(name)
      ...
    END_USE_CODE    
 */
case 12:
YY_RULE_SETUP
#line 294 "scanner.l"
{
    //printf("{%s}\n", yytext);
    yy_push_state(PREPROC_IF, yyscanner);
}
	YY_BREAK
case 13:
YY_RULE_SETUP
#line 298 "scanner.l"
{
    //printf("[EAT{%s}\n", yytext);
    yy_push_state(PREPROC_BODY_EAT, yyscanner);
}
	YY_BREAK
case 14:
YY_RULE_SETUP
#line 302 "scanner.l"
{
    //printf("USE_CODE_IF_NOT\n");
    yy_push_state(PREPROC_IF_NOT, yyscanner);
}
	YY_BREAK
case 15:
YY_RULE_SETUP
#line 306 "scanner.l"
{
    //printf("[EAT{%s}\n", yytext);
    yy_push_state(PREPROC_BODY_EAT, yyscanner);
}
	YY_BREAK
case 16:
YY_RULE_SETUP
#line 310 "scanner.l"
{
    //printf("preproc use code if '%s'\n", yytext);
    yy_pop_state(yyscanner);
    if (yyextra->isPreprocessorConditionSet(yytext))
        yy_push_state(PREPROC_PRE_BODY_USE, yyscanner);
    else
        yy_push_state(PREPROC_PRE_BODY_EAT, yyscanner);
}
	YY_BREAK
case 17:
YY_RULE_SETUP
#line 318 "scanner.l"
{
    //printf("preproc use code if not '%s'\n", yytext);
    yy_pop_state(yyscanner);
    if (!yyextra->isPreprocessorConditionSet(yytext))
        yy_push_state(PREPROC_PRE_BODY_USE, yyscanner);
    else
        yy_push_state(PREPROC_PRE_BODY_EAT, yyscanner);
}
	YY_BREAK
case 18:
YY_RULE_SETUP
#line 326 "scanner.l"
{
    yy_pop_state(yyscanner);
    yy_push_state(PREPROC_BODY_USE, yyscanner);
}
	YY_BREAK
case 19:
YY_RULE_SETUP
#line 330 "scanner.l"
{
    //printf("PREPROCESSOR EAT : {%s}\n", yytext);
    yy_pop_state(yyscanner);
    yy_push_state(PREPROC_BODY_EAT, yyscanner);
}
	YY_BREAK
case 20:
YY_RULE_SETUP
#line 335 "scanner.l"
{
    //printf("-->END_USE_CODE\n");
    yy_pop_state(yyscanner);
}
	YY_BREAK
case 21:
/* rule 21 can match eol */
YY_RULE_SETUP
#line 339 "scanner.l"
{ /* eat up code block filtered out by preprocessor */
    //printf("PREPROCESSOR EAT2 : {%s}\n", yytext);
    processLocation();
}
	YY_BREAK
case 22:
YY_RULE_SETUP
#line 343 "scanner.l"
{ /* eat up code block filtered out by preprocessor */
    //printf("PREPROCESSOR EAT3 : {%s}\n", yytext);
    yyextra->addPreprocessorComment(yylloc->first_line, yylloc->last_line,
                                    yylloc->first_column+1, yylloc->last_column+1,
                                    yylloc->first_byte, yylloc->length_bytes);
}
	YY_BREAK
/* Language keywords */
case 23:
YY_RULE_SETUP
#line 353 "scanner.l"
return ON;
	YY_BREAK
case 24:
YY_RULE_SETUP
#line 354 "scanner.l"
return END;
	YY_BREAK
case 25:
YY_RULE_SETUP
#line 355 "scanner.l"
return NOTE;
	YY_BREAK
case 26:
YY_RULE_SETUP
#line 356 "scanner.l"
return INIT;
	YY_BREAK
case 27:
YY_RULE_SETUP
#line 357 "scanner.l"
return DECLARE;
	YY_BREAK
case 28:
YY_RULE_SETUP
#line 358 "scanner.l"
return WHILE;
	YY_BREAK
case 29:
YY_RULE_SETUP
#line 359 "scanner.l"
return IF;
	YY_BREAK
case 30:
YY_RULE_SETUP
#line 360 "scanner.l"
return BITWISE_OR;
	YY_BREAK
case 31:
YY_RULE_SETUP
#line 361 "scanner.l"
return OR;
	YY_BREAK
case 32:
YY_RULE_SETUP
#line 362 "scanner.l"
return RELEASE;
	YY_BREAK
case 33:
YY_RULE_SETUP
#line 363 "scanner.l"
return BITWISE_AND;
	YY_BREAK
case 34:
YY_RULE_SETUP
#line 364 "scanner.l"
return AND;
	YY_BREAK
case 35:
YY_RULE_SETUP
#line 365 "scanner.l"
return BITWISE_NOT;
	YY_BREAK
case 36:
YY_RULE_SETUP
#line 366 "scanner.l"
return NOT;
	YY_BREAK
case 37:
YY_RULE_SETUP
#line 367 "scanner.l"
return ELSE;
	YY_BREAK
case 38:
YY_RULE_SETUP
#line 368 "scanner.l"
return CONTROLLER;
	YY_BREAK
case 39:
YY_RULE_SETUP
#line 369 "scanner.l"
return RPN;
	YY_BREAK
case 40:
YY_RULE_SETUP
#line 370 "scanner.l"
return NRPN;
	YY_BREAK
case 41:
YY_RULE_SETUP
#line 371 "scanner.l"
return CASE;
	YY_BREAK
case 42:
YY_RULE_SETUP
#line 372 "scanner.l"
return SELECT;
	YY_BREAK
case 43:
YY_RULE_SETUP
#line 373 "scanner.l"
return TO;
	YY_BREAK
case 44:
YY_RULE_SETUP
#line 374 "scanner.l"
return LE;
	YY_BREAK
case 45:
YY_RULE_SETUP
#line 375 "scanner.l"
return GE;
	YY_BREAK
case 46:
YY_RULE_SETUP
#line 376 "scanner.l"
return RIGHT_ARROW;
	YY_BREAK
case 47:
YY_RULE_SETUP
#line 377 "scanner.l"
return CONST_; // note: "CONST" is already defined for C/C++ compilers on Windows by default
	YY_BREAK
case 48:
YY_RULE_SETUP
#line 378 "scanner.l"
return POLYPHONIC;
	YY_BREAK
case 49:
YY_RULE_SETUP
#line 379 "scanner.l"
return PATCH;
	YY_BREAK
case 50:
YY_RULE_SETUP
#line 380 "scanner.l"
return LOCAL;
	YY_BREAK
case 51:
YY_RULE_SETUP
#line 381 "scanner.l"
return MOD;
	YY_BREAK
case 52:
YY_RULE_SETUP
#line 382 "scanner.l"
return FUNCTION;
	YY_BREAK
case 53:
YY_RULE_SETUP
#line 383 "scanner.l"
return CALL;
	YY_BREAK
case 54:
YY_RULE_SETUP
#line 384 "scanner.l"
return SYNCHRONIZED;
	YY_BREAK
case 55:
YY_RULE_SETUP
#line 386 "scanner.l"
{
    return *yytext;
}
	YY_BREAK
case 56:
YY_RULE_SETUP
#line 390 "scanner.l"
{
    yylval->sValue = strdup(yytext);
    yyextra->autoFreeAfterParse(yylval->sValue);
    return VARIABLE;
}
	YY_BREAK
case 57:
YY_RULE_SETUP
#line 396 "scanner.l"
{
    yylval->sValue = strdup(yytext);
    yyextra->autoFreeAfterParse(yylval->sValue);
    return IDENTIFIER;
}
	YY_BREAK
case 58:
YY_RULE_SETUP
#line 402 "scanner.l"
return ASSIGNMENT;
	YY_BREAK
case 59:
/* rule 59 can match eol */
YY_RULE_SETUP
#line 404 "scanner.l"
{
     yylineno += countNewLineChars(yytext);
     yycolumn = 0;
    //printf("lex: new line %d\n", yylineno, yytext);
    //return LF;
}
	YY_BREAK
case 60:
/* rule 60 can match eol */
YY_RULE_SETUP
#line 411 "scanner.l"
{ /* eat up comments */
    processLocation();
}
	YY_BREAK
case 61:
YY_RULE_SETUP
#line 415 "scanner.l"
/* eat up whitespace */
	YY_BREAK
case 62:
YY_RULE_SETUP
#line 417 "scanner.l"
{
    printf( "Unrecognized character: '%s' (line %d, column %d)\n", yytext, yylineno, yycolumn);
    return UNKNOWN_CHAR;
}
	YY_BREAK
case 63:
YY_RULE_SETUP
#line 422 "scanner.l"
ECHO;
	YY_BREAK
#line 2084 "scanner.cpp"
case YY_STATE_EOF(INITIAL):
case YY_STATE_EOF(PREPROC_BODY_USE):
case YY_STATE_EOF(PREPROC_SET_COND):
case YY_STATE_EOF(PREPROC_RESET_COND):
case YY_STATE_EOF(PREPROC_IF):
case YY_STATE_EOF(PREPROC_IF_NOT):
case YY_STATE_EOF(PREPROC_BODY_EAT):
case YY_STATE_EOF(PREPROC_PRE_BODY_USE):
case YY_STATE_EOF(PREPROC_PRE_BODY_EAT):
	yyterminate();

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - yyg->yytext_ptr) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = yyg->yy_hold_char;
		YY_RESTORE_YY_MORE_OFFSET

		if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed yyin at a new source and called
			 * yylex().  If so, then we have to assure
			 * consistency between YY_CURRENT_BUFFER and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
			YY_CURRENT_BUFFER_LVALUE->yy_input_file = yyin;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( yyg->yy_c_buf_p <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			yyg->yy_c_buf_p = yyg->yytext_ptr + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state( yyscanner );

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state , yyscanner);

			yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++yyg->yy_c_buf_p;
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
				yy_cp = yyg->yy_c_buf_p;
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer( yyscanner ) )
			{
			case EOB_ACT_END_OF_FILE:
				{
				yyg->yy_did_buffer_switch_on_eof = 0;

				if ( yywrap( yyscanner ) )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * yytext, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					yyg->yy_c_buf_p = yyg->yytext_ptr + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				yyg->yy_c_buf_p =
					yyg->yytext_ptr + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				yyg->yy_c_buf_p =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars];

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
	} /* end of user's declarations */
} /* end of yylex */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
static int yy_get_next_buffer (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
	char *source = yyg->yytext_ptr;
	int number_to_move, i;
	int ret_val;

	if ( yyg->yy_c_buf_p > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( yyg->yy_c_buf_p - yyg->yytext_ptr - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr - 1);

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars = 0;

	else
		{
			int num_to_read =
			YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = YY_CURRENT_BUFFER_LVALUE;

			int yy_c_buf_p_offset =
				(int) (yyg->yy_c_buf_p - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				int new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					yyrealloc( (void *) b->yy_ch_buf,
							 (yy_size_t) (b->yy_buf_size + 2) , yyscanner );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = NULL;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			yyg->yy_c_buf_p = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size -
						number_to_move - 1;

			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]),
			yyg->yy_n_chars, num_to_read );

		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	if ( yyg->yy_n_chars == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			yyrestart( yyin  , yyscanner);
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	if ((yyg->yy_n_chars + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size) {
		/* Extend the array by 50%, plus the number we really need. */
		int new_size = yyg->yy_n_chars + number_to_move + (yyg->yy_n_chars >> 1);
		YY_CURRENT_BUFFER_LVALUE->yy_ch_buf = (char *) yyrealloc(
			(void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf, (yy_size_t) new_size , yyscanner );
		if ( ! YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			YY_FATAL_ERROR( "out of dynamic memory in yy_get_next_buffer()" );
		/* "- 2" to take care of EOB's */
		YY_CURRENT_BUFFER_LVALUE->yy_buf_size = (int) (new_size - 2);
	}

	yyg->yy_n_chars += number_to_move;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] = YY_END_OF_BUFFER_CHAR;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;

	yyg->yytext_ptr = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

	return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

    static yy_state_type yy_get_previous_state (yyscan_t yyscanner)
{
	yy_state_type yy_current_state;
	char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yy_current_state = yyg->yy_start;

	for ( yy_cp = yyg->yytext_ptr + YY_MORE_ADJ; yy_cp < yyg->yy_c_buf_p; ++yy_cp )
		{
		YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 1);
		if ( yy_accept[yy_current_state] )
			{
			yyg->yy_last_accepting_state = yy_current_state;
			yyg->yy_last_accepting_cpos = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 563 )
				yy_c = yy_meta[yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
		}

	return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
    static yy_state_type yy_try_NUL_trans  (yy_state_type yy_current_state , yyscan_t yyscanner)
{
	int yy_is_jam;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner; /* This var may be unused depending upon options. */
	char *yy_cp = yyg->yy_c_buf_p;

	YY_CHAR yy_c = 1;
	if ( yy_accept[yy_current_state] )
		{
		yyg->yy_last_accepting_state = yy_current_state;
		yyg->yy_last_accepting_cpos = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 563 )
			yy_c = yy_meta[yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
	yy_is_jam = (yy_current_state == 562);

	(void)yyg;
	return yy_is_jam ? 0 : yy_current_state;
}

#ifndef YY_NO_UNPUT

    static void yyunput (int c, char * yy_bp , yyscan_t yyscanner)
{
	char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    yy_cp = yyg->yy_c_buf_p;

	/* undo effects of setting up yytext */
	*yy_cp = yyg->yy_hold_char;

	if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
		{ /* need to shift things up to make room */
		/* +2 for EOB chars. */
		int number_to_move = yyg->yy_n_chars + 2;
		char *dest = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[
					YY_CURRENT_BUFFER_LVALUE->yy_buf_size + 2];
		char *source =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move];

		while ( source > YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			*--dest = *--source;

		yy_cp += (int) (dest - source);
		yy_bp += (int) (dest - source);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars =
			yyg->yy_n_chars = (int) YY_CURRENT_BUFFER_LVALUE->yy_buf_size;

		if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
			YY_FATAL_ERROR( "flex scanner push-back overflow" );
		}

	*--yy_cp = (char) c;

	yyg->yytext_ptr = yy_bp;
	yyg->yy_hold_char = *yy_cp;
	yyg->yy_c_buf_p = yy_cp;
}

#endif

#ifndef YY_NO_INPUT
#ifdef __cplusplus
    static int yyinput (yyscan_t yyscanner)
#else
    static int input  (yyscan_t yyscanner)
#endif

{
	int c;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	*yyg->yy_c_buf_p = yyg->yy_hold_char;

	if ( *yyg->yy_c_buf_p == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( yyg->yy_c_buf_p < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			/* This was really a NUL. */
			*yyg->yy_c_buf_p = '\0';

		else
			{ /* need more input */
			int offset = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr);
			++yyg->yy_c_buf_p;

			switch ( yy_get_next_buffer( yyscanner ) )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					yyrestart( yyin , yyscanner);

					/*FALLTHROUGH*/

				case EOB_ACT_END_OF_FILE:
					{
					if ( yywrap( yyscanner ) )
						return 0;

					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput(yyscanner);
#else
					return input(yyscanner);
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					yyg->yy_c_buf_p = yyg->yytext_ptr + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) yyg->yy_c_buf_p;	/* cast for 8-bit char's */
	*yyg->yy_c_buf_p = '\0';	/* preserve yytext */
	yyg->yy_hold_char = *++yyg->yy_c_buf_p;

	return c;
}
#endif	/* ifndef YY_NO_INPUT */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * @param yyscanner The scanner object.
 * @note This function does not reset the start condition to @c INITIAL .
 */
    void yyrestart  (FILE * input_file , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! YY_CURRENT_BUFFER ){
        yyensure_buffer_stack (yyscanner);
		YY_CURRENT_BUFFER_LVALUE =
            yy_create_buffer( yyin, YY_BUF_SIZE , yyscanner);
	}

	yy_init_buffer( YY_CURRENT_BUFFER, input_file , yyscanner);
	yy_load_buffer_state( yyscanner );
}

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * @param yyscanner The scanner object.
 */
    void yy_switch_to_buffer  (YY_BUFFER_STATE  new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	/* TODO. We should be able to replace this entire function body
	 * with
	 *		yypop_buffer_state();
	 *		yypush_buffer_state(new_buffer);
     */
	yyensure_buffer_stack (yyscanner);
	if ( YY_CURRENT_BUFFER == new_buffer )
		return;

	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	YY_CURRENT_BUFFER_LVALUE = new_buffer;
	yy_load_buffer_state( yyscanner );

	/* We don't actually know whether we did this switch during
	 * EOF (yywrap()) processing, but the only time this flag
	 * is looked at is after yywrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	yyg->yy_did_buffer_switch_on_eof = 1;
}

static void yy_load_buffer_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
	yyg->yytext_ptr = yyg->yy_c_buf_p = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
	yyin = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
	yyg->yy_hold_char = *yyg->yy_c_buf_p;
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * @param yyscanner The scanner object.
 * @return the allocated buffer state.
 */
    YY_BUFFER_STATE yy_create_buffer  (FILE * file, int  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	b = (YY_BUFFER_STATE) yyalloc( sizeof( struct yy_buffer_state ) , yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) yyalloc( (yy_size_t) (b->yy_buf_size + 2) , yyscanner );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_is_our_buffer = 1;

	yy_init_buffer( b, file , yyscanner);

	return b;
}

/** Destroy the buffer.
 * @param b a buffer created with yy_create_buffer()
 * @param yyscanner The scanner object.
 */
    void yy_delete_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! b )
		return;

	if ( b == YY_CURRENT_BUFFER ) /* Not sure if we should pop here. */
		YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) 0;

	if ( b->yy_is_our_buffer )
		yyfree( (void *) b->yy_ch_buf , yyscanner );

	yyfree( (void *) b , yyscanner );
}

/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a yyrestart() or at EOF.
 */
    static void yy_init_buffer  (YY_BUFFER_STATE  b, FILE * file , yyscan_t yyscanner)

{
	int oerrno = errno;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yy_flush_buffer( b , yyscanner);

	b->yy_input_file = file;
	b->yy_fill_buffer = 1;

    /* If b is the current buffer, then yy_init_buffer was _probably_
     * called from yyrestart() or through yy_get_next_buffer.
     * In that case, we don't want to reset the lineno or column.
     */
    if (b != YY_CURRENT_BUFFER){
        b->yy_bs_lineno = 1;
        b->yy_bs_column = 0;
    }

        b->yy_is_interactive = file ? (isatty( fileno(file) ) > 0) : 0;
    
	errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * @param yyscanner The scanner object.
 */
    void yy_flush_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == YY_CURRENT_BUFFER )
		yy_load_buffer_state( yyscanner );
}

/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  @param yyscanner The scanner object.
 */
void yypush_buffer_state (YY_BUFFER_STATE new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (new_buffer == NULL)
		return;

	yyensure_buffer_stack(yyscanner);

	/* This block is copied from yy_switch_to_buffer. */
	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	/* Only push if top exists. Otherwise, replace top. */
	if (YY_CURRENT_BUFFER)
		yyg->yy_buffer_stack_top++;
	YY_CURRENT_BUFFER_LVALUE = new_buffer;

	/* copied from yy_switch_to_buffer. */
	yy_load_buffer_state( yyscanner );
	yyg->yy_did_buffer_switch_on_eof = 1;
}

/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  @param yyscanner The scanner object.
 */
void yypop_buffer_state (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (!YY_CURRENT_BUFFER)
		return;

	yy_delete_buffer(YY_CURRENT_BUFFER , yyscanner);
	YY_CURRENT_BUFFER_LVALUE = NULL;
	if (yyg->yy_buffer_stack_top > 0)
		--yyg->yy_buffer_stack_top;

	if (YY_CURRENT_BUFFER) {
		yy_load_buffer_state( yyscanner );
		yyg->yy_did_buffer_switch_on_eof = 1;
	}
}

/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
static void yyensure_buffer_stack (yyscan_t yyscanner)
{
	yy_size_t num_to_alloc;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if (!yyg->yy_buffer_stack) {

		/* First allocation is just for 2 elements, since we don't know if this
		 * scanner will even need a stack. We use 2 instead of 1 to avoid an
		 * immediate realloc on the next call.
         */
      num_to_alloc = 1; /* After all that talk, this was set to 1 anyways... */
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yyalloc
								(num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );

		memset(yyg->yy_buffer_stack, 0, num_to_alloc * sizeof(struct yy_buffer_state*));

		yyg->yy_buffer_stack_max = num_to_alloc;
		yyg->yy_buffer_stack_top = 0;
		return;
	}

	if (yyg->yy_buffer_stack_top >= (yyg->yy_buffer_stack_max) - 1){

		/* Increase the buffer to prepare for a possible push. */
		yy_size_t grow_size = 8 /* arbitrary grow size */;

		num_to_alloc = yyg->yy_buffer_stack_max + grow_size;
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yyrealloc
								(yyg->yy_buffer_stack,
								num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );

		/* zero only the new slots.*/
		memset(yyg->yy_buffer_stack + yyg->yy_buffer_stack_max, 0, grow_size * sizeof(struct yy_buffer_state*));
		yyg->yy_buffer_stack_max = num_to_alloc;
	}
}

/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yy_scan_buffer  (char * base, yy_size_t  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	if ( size < 2 ||
	     base[size-2] != YY_END_OF_BUFFER_CHAR ||
	     base[size-1] != YY_END_OF_BUFFER_CHAR )
		/* They forgot to leave room for the EOB's. */
		return NULL;

	b = (YY_BUFFER_STATE) yyalloc( sizeof( struct yy_buffer_state ) , yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_buffer()" );

	b->yy_buf_size = (int) (size - 2);	/* "- 2" to take care of EOB's */
	b->yy_buf_pos = b->yy_ch_buf = base;
	b->yy_is_our_buffer = 0;
	b->yy_input_file = NULL;
	b->yy_n_chars = b->yy_buf_size;
	b->yy_is_interactive = 0;
	b->yy_at_bol = 1;
	b->yy_fill_buffer = 0;
	b->yy_buffer_status = YY_BUFFER_NEW;

	yy_switch_to_buffer( b , yyscanner );

	return b;
}

/** Setup the input buffer state to scan a string. The next call to yylex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       yy_scan_bytes() instead.
 */
YY_BUFFER_STATE yy_scan_string (const char * yystr , yyscan_t yyscanner)
{
    
	return yy_scan_bytes( yystr, (int) strlen(yystr) , yyscanner);
}

/** Setup the input buffer state to scan the given bytes. The next call to yylex() will
 * scan from a @e copy of @a bytes.
 * @param yybytes the byte buffer to scan
 * @param _yybytes_len the number of bytes in the buffer pointed to by @a bytes.
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yy_scan_bytes  (const char * yybytes, int  _yybytes_len , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
	char *buf;
	yy_size_t n;
	int i;
    
	/* Get memory for full buffer, including space for trailing EOB's. */
	n = (yy_size_t) (_yybytes_len + 2);
	buf = (char *) yyalloc( n , yyscanner );
	if ( ! buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_bytes()" );

	for ( i = 0; i < _yybytes_len; ++i )
		buf[i] = yybytes[i];

	buf[_yybytes_len] = buf[_yybytes_len+1] = YY_END_OF_BUFFER_CHAR;

	b = yy_scan_buffer( buf, n , yyscanner);
	if ( ! b )
		YY_FATAL_ERROR( "bad buffer in yy_scan_bytes()" );

	/* It's okay to grow etc. this buffer, and we should throw it
	 * away when we're done.
	 */
	b->yy_is_our_buffer = 1;

	return b;
}

    static void yy_push_state (int  _new_state , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( yyg->yy_start_stack_ptr >= yyg->yy_start_stack_depth )
		{
		yy_size_t new_size;

		yyg->yy_start_stack_depth += YY_START_STACK_INCR;
		new_size = (yy_size_t) yyg->yy_start_stack_depth * sizeof( int );

		if ( ! yyg->yy_start_stack )
			yyg->yy_start_stack = (int *) yyalloc( new_size , yyscanner );

		else
			yyg->yy_start_stack = (int *) yyrealloc(
					(void *) yyg->yy_start_stack, new_size , yyscanner );

		if ( ! yyg->yy_start_stack )
			YY_FATAL_ERROR( "out of memory expanding start-condition stack" );
		}

	yyg->yy_start_stack[yyg->yy_start_stack_ptr++] = YY_START;

	BEGIN(_new_state);
}

    static void yy_pop_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( --yyg->yy_start_stack_ptr < 0 )
		YY_FATAL_ERROR( "start-condition stack underflow" );

	BEGIN(yyg->yy_start_stack[yyg->yy_start_stack_ptr]);
}

    static int yy_top_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	return yyg->yy_start_stack[yyg->yy_start_stack_ptr - 1];
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

static void yynoreturn yy_fatal_error (const char* msg , yyscan_t yyscanner)
{
	struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	(void)yyg;
	fprintf( stderr, "%s\n", msg );
	exit( YY_EXIT_FAILURE );
}

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		yytext[yyleng] = yyg->yy_hold_char; \
		yyg->yy_c_buf_p = yytext + yyless_macro_arg; \
		yyg->yy_hold_char = *yyg->yy_c_buf_p; \
		*yyg->yy_c_buf_p = '\0'; \
		yyleng = yyless_macro_arg; \
		} \
	while ( 0 )

/* Accessor  methods (get/set functions) to struct members. */

/** Get the user-defined data for this scanner.
 * @param yyscanner The scanner object.
 */
YY_EXTRA_TYPE yyget_extra  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyextra;
}

/** Get the current line number.
 * @param yyscanner The scanner object.
 */
int yyget_lineno  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yylineno;
}

/** Get the current column number.
 * @param yyscanner The scanner object.
 */
int yyget_column  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yycolumn;
}

/** Get the input stream.
 * @param yyscanner The scanner object.
 */
FILE *yyget_in  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyin;
}

/** Get the output stream.
 * @param yyscanner The scanner object.
 */
FILE *yyget_out  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyout;
}

/** Get the length of the current token.
 * @param yyscanner The scanner object.
 */
int yyget_leng  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyleng;
}

/** Get the current token.
 * @param yyscanner The scanner object.
 */

char *yyget_text  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yytext;
}

/** Set the user-defined data. This data is never touched by the scanner.
 * @param user_defined The data to be associated with this scanner.
 * @param yyscanner The scanner object.
 */
void yyset_extra (YY_EXTRA_TYPE  user_defined , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyextra = user_defined ;
}

/** Set the current line number.
 * @param _line_number line number
 * @param yyscanner The scanner object.
 */
void yyset_lineno (int  _line_number , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* lineno is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           YY_FATAL_ERROR( "yyset_lineno called with no buffer" );
    
    yylineno = _line_number;
}

/** Set the current column.
 * @param _column_no column number
 * @param yyscanner The scanner object.
 */
void yyset_column (int  _column_no , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* column is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           YY_FATAL_ERROR( "yyset_column called with no buffer" );
    
    yycolumn = _column_no;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param _in_str A readable stream.
 * @param yyscanner The scanner object.
 * @see yy_switch_to_buffer
 */
void yyset_in (FILE *  _in_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyin = _in_str ;
}

void yyset_out (FILE *  _out_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyout = _out_str ;
}

int yyget_debug  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yy_flex_debug;
}

void yyset_debug (int  _bdebug , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yy_flex_debug = _bdebug ;
}

/* Accessor methods for yylval and yylloc */

YYSTYPE * yyget_lval  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yylval;
}

void yyset_lval (YYSTYPE *  yylval_param , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yylval = yylval_param;
}

YYLTYPE *yyget_lloc  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yylloc;
}
    
void yyset_lloc (YYLTYPE *  yylloc_param , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yylloc = yylloc_param;
}
    
/* User-visible API */

/* yylex_init is special because it creates the scanner itself, so it is
 * the ONLY reentrant function that doesn't take the scanner as the last argument.
 * That's why we explicitly handle the declaration, instead of using our macros.
 */
int yylex_init(yyscan_t* ptr_yy_globals)
{
    if (ptr_yy_globals == NULL){
        errno = EINVAL;
        return 1;
    }

    *ptr_yy_globals = (yyscan_t) yyalloc ( sizeof( struct yyguts_t ), NULL );

    if (*ptr_yy_globals == NULL){
        errno = ENOMEM;
        return 1;
    }

    /* By setting to 0xAA, we expose bugs in yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));

    return yy_init_globals ( *ptr_yy_globals );
}

/* yylex_init_extra has the same functionality as yylex_init, but follows the
 * convention of taking the scanner as the last argument. Note however, that
 * this is a *pointer* to a scanner, as it will be allocated by this call (and
 * is the reason, too, why this function also must handle its own declaration).
 * The user defined value in the first argument will be available to yyalloc in
 * the yyextra field.
 */
int yylex_init_extra( YY_EXTRA_TYPE yy_user_defined, yyscan_t* ptr_yy_globals )
{
    struct yyguts_t dummy_yyguts;

    yyset_extra (yy_user_defined, &dummy_yyguts);

    if (ptr_yy_globals == NULL){
        errno = EINVAL;
        return 1;
    }

    *ptr_yy_globals = (yyscan_t) yyalloc ( sizeof( struct yyguts_t ), &dummy_yyguts );

    if (*ptr_yy_globals == NULL){
        errno = ENOMEM;
        return 1;
    }

    /* By setting to 0xAA, we expose bugs in
    yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));

    yyset_extra (yy_user_defined, *ptr_yy_globals);

    return yy_init_globals ( *ptr_yy_globals );
}

static int yy_init_globals (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    /* Initialization is the same as for the non-reentrant scanner.
     * This function is called from yylex_destroy(), so don't allocate here.
     */

    yyg->yy_buffer_stack = NULL;
    yyg->yy_buffer_stack_top = 0;
    yyg->yy_buffer_stack_max = 0;
    yyg->yy_c_buf_p = NULL;
    yyg->yy_init = 0;
    yyg->yy_start = 0;

    yyg->yy_start_stack_ptr = 0;
    yyg->yy_start_stack_depth = 0;
    yyg->yy_start_stack =  NULL;

/* Defined in main.c */
#ifdef YY_STDINIT
    yyin = stdin;
    yyout = stdout;
#else
    yyin = NULL;
    yyout = NULL;
#endif

    /* For future reference: Set errno on error, since we are called by
     * yylex_init()
     */
    return 0;
}

/* yylex_destroy is for both reentrant and non-reentrant scanners. */
int yylex_destroy  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    /* Pop the buffer stack, destroying each element. */
	while(YY_CURRENT_BUFFER){
		yy_delete_buffer( YY_CURRENT_BUFFER , yyscanner );
		YY_CURRENT_BUFFER_LVALUE = NULL;
		yypop_buffer_state(yyscanner);
	}

	/* Destroy the stack itself. */
	yyfree(yyg->yy_buffer_stack , yyscanner);
	yyg->yy_buffer_stack = NULL;

    /* Destroy the start condition stack. */
        yyfree( yyg->yy_start_stack , yyscanner );
        yyg->yy_start_stack = NULL;

    /* Reset the globals. This is important in a non-reentrant scanner so the next time
     * yylex() is called, initialization will occur. */
    yy_init_globals( yyscanner);

    /* Destroy the main struct (reentrant only). */
    yyfree ( yyscanner , yyscanner );
    yyscanner = NULL;
    return 0;
}

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void yy_flex_strncpy (char* s1, const char * s2, int n , yyscan_t yyscanner)
{
	struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	(void)yyg;

	int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (const char * s , yyscan_t yyscanner)
{
	int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
}
#endif

void *yyalloc (yy_size_t  size , yyscan_t yyscanner)
{
	struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	(void)yyg;
	return malloc(size);
}

void *yyrealloc  (void * ptr, yy_size_t  size , yyscan_t yyscanner)
{
	struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	(void)yyg;

	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
	return realloc(ptr, size);
}

void yyfree (void * ptr , yyscan_t yyscanner)
{
	struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	(void)yyg;
	free( (char *) ptr );	/* see yyrealloc() for (char *) cast */
}

#define YYTABLES_NAME "yytables"

#line 422 "scanner.l"


namespace LinuxSampler {

void ParserContext::createScanner(std::istream* is) {
   if (scanner) destroyScanner();
   this->is = is;
   yylex_init(&scanner);
   yyset_extra(this, scanner);
}

void ParserContext::destroyScanner() {
   if (!scanner) return;
   yylex_destroy(scanner);
   scanner = NULL;
}

} // namespace LinuxSampler

